import os
import sys
import hydra
import platform
import threading
import subprocess
from time import sleep
from typing import List, Tuple
from loguru import logger
from omegaconf import OmegaConf

import idlem.utils as ut
from idlem.arg_parser import Config

DOCKER_NAME = "test"

# setup logger
logger.configure(handlers=[{"sink": sys.stdout, "format": "{time:[MM-DD HH:mm:ss]} - {message}"}])


class MinerThread(threading.Thread):
    def __init__(self, start_command: str):
        super().__init__()
        self.start_command = start_command

    def run(self):
        subprocess.run(self.start_command, shell=True, stdout=subprocess.DEVNULL)

    def stop(self):
        subprocess.run(f"docker kill {DOCKER_NAME}", shell=True, stdout=subprocess.DEVNULL)


class Miner:
    """High-level wrapper which handles restaring the theads"""

    def __init__(
        self,
        scheduler: List[Tuple[str, str]],
        allowed_users: List[str],
    ):
        self.scheduler = scheduler
        self.miner_thread = None
        self.start_free_gpus_set = {}
        self.allowed_users = allowed_users

        # get params str with `devices` being filled later. `python3` here is renamed miner!
        self.params_str = (
            f"""docker run --name {DOCKER_NAME} --rm -it """
            + """--gpus '\"device={devices}\"' """
            + f"""-v {os.getcwd()}:/usr/bin/logs/ -v {ut.MINER_CONFIG_PATH}:/usr/bin/config.txt gm ./python3 --config ./config.txt"""
        )

    def step(self):
        on_schedule = ut.is_in_time_range(self.scheduler)
        extra_users = ut.get_external_users(self.allowed_users)
        if extra_users:
            logger.info(f"Someone logged in: {extra_users}. Stoping.")

        # stop when someone logs in or because of scheduler
        need_stop = not on_schedule or extra_users

        new_free_gpus = set(ut.free_gpus_list()).difference(self.start_free_gpus_set)
        have_new_free_gpus = len(new_free_gpus) > 0
        # restart when can run and miner is off (or it's on, but we can grab more GPUs)
        if not need_stop and (have_new_free_gpus or not self.working):
            self.restart()

        if need_stop:
            self.stop()

        sleep(10)

    def start(self):
        self.start_free_gpus_set = set(ut.free_gpus_list())
        logger.info(f"Started mining. Using GPUs: {self.start_free_gpus_set}")
        self.miner_thread = MinerThread(self.params_str.format(devices=self.free_gpus_str))
        self.miner_thread.start()
        sleep(10)  # extra time for startup

    def stop(self):
        if not self.working:
            return
        logger.info(f"Stopped mining on gpus: {self.start_free_gpus_set}")
        self.start_free_gpus_set = {}
        self.miner_thread = self.miner_thread.stop()
        self.miner_thread = None
        sleep(10)  # make sure all processes stop

    def restart(self):
        self.stop()
        self.start()

    @property
    def num_free_gpus(self) -> int:
        return len(ut.free_gpus_list())

    @property
    def free_gpus_str(self) -> str:
        free_gpus_list = ut.free_gpus_list()
        return ",".join(free_gpus_list) if free_gpus_list else None

    @property
    def on_schedule(self):
        return ut.is_in_time_range(self.schedule_dict)

    @property
    def working(self):
        return self.miner_thread is not None


@hydra.main(config_path="./configs/", config_name="config.yaml")
def main(cfg: Config) -> None:
    logger.info(OmegaConf.to_yaml(cfg, resolve=True))

    cfg.miner.server.user += f".{platform.node()}"  # set unique name for each node
    # dump config to tmp. later this file would be mounted to docker
    ut.dump_gminer_config(cfg.miner, ut.MINER_CONFIG_PATH)

    miner = Miner(cfg.scheduler.data, cfg.allowed_users)
    while True:
        miner.step()


if __name__ == "__main__":
    main()



"""script to be run on Zhores cluster"""
from logging import log
import os
import sys
import nvsmi
import hydra
import platform
import threading
import subprocess
from time import sleep
from typing import List
from loguru import logger
from omegaconf import OmegaConf

import idlem.utils as ut
from idlem.arg_parser import ConfigSlurm

# setup logger
logger.configure(handlers=[{"sink": sys.stdout, "format": "{time:[MM-DD HH:mm:ss]} - {message}"}])


@hydra.main(config_path="./configs/", config_name="config_slurm.yaml")
def main(cfg: ConfigSlurm):
    logger.info(OmegaConf.to_yaml(cfg, resolve=True))
    # hydra changes dir. need to undo this
    os.chdir(hydra.utils.get_original_cwd())

    cfg.miner.server.user += f".{platform.node()}"  # set unique name for each node

    while True:
        num_running = ut.slurm_num_running()
        on_schedule = ut.is_in_time_range(cfg.scheduler.data)
        log_str = f"Current mining jobs: {num_running}/{cfg.max_schedule_jobs if on_schedule else cfg.max_idle_jobs}. "
        logger.info(log_str + f"On schedule: {'True' if on_schedule else 'False'}")

        if on_schedule:
            if num_running < cfg.max_schedule_jobs:
                logger.info("spawning more miners")
                ut.slurm_submit_mining_jobs(num_jobs=cfg.max_schedule_jobs - num_running)

        if not on_schedule:
            if num_running > cfg.max_idle_jobs:
                logger.info("Too many miners not on schedule. Killing")
                ut.slurm_kill_all()
                sleep(10)  # give jobs enough time to die
                num_running = ut.slurm_num_running()

            if num_running < cfg.max_idle_jobs:
                logger.info("Spawning more miners")
                ut.slurm_submit_mining_jobs(num_jobs=cfg.max_idle_jobs - num_running)

        sleep(cfg.sleep_time)


if __name__ == "__main__":
    main()



import sys
from time import sleep
from omegaconf import OmegaConf
from freezegun import freeze_time
from datetimerange import DateTimeRange


sys.path.append("src")
import utils


def test_str_to_weekday():
    assert utils.str_to_weekday("Mon") == 1
    assert utils.str_to_weekday("Sun") == 7


FROZEN_DATE = "2021-07-01"


@freeze_time(FROZEN_DATE)
def test_parse_time_range():
    # test parsing single range
    time_ranges = "00:00-23:59"
    parsed = utils.parse_time_range(time_ranges)
    expected_parsed = [DateTimeRange(f"{FROZEN_DATE} 00:00:00+0", f"{FROZEN_DATE} 23:59:00+0")]
    assert parsed == expected_parsed

    # test parsing multiple ranges
    time_ranges = "00:02-10:05,21:02-23:59"
    parsed = utils.parse_time_range(time_ranges)
    expected_parsed = [
        DateTimeRange(f"{FROZEN_DATE} 00:02:00+0", f"{FROZEN_DATE} 10:05:00+0"),
        DateTimeRange(f"{FROZEN_DATE} 21:02:00+0", f"{FROZEN_DATE} 23:59:00+0"),
    ]
    assert parsed == expected_parsed


@freeze_time("2021-07-01T23:59:59", tick=True, tz_offset=-3)
def test_is_in_timerange():
    scheduler = [["Thu", "00:00-23:59"], ["Fri", "00:00-23:59"]]
    assert utils.true_now().isoweekday() == utils.str_to_weekday("Thu")
    assert not utils.is_in_time_range(scheduler)

    sleep(1)  # wait for day change
    assert utils.true_now().isoweekday() == utils.str_to_weekday("Fri")
    assert utils.is_in_time_range(scheduler)


def test_dump_config():
    """check that dumping works and produces expected format for configs"""
    cfg = OmegaConf.create({"common": {"algo": "eth", "interval": 60}, "server": {"host": "some.host", "port": 42}})
    TEST_CFG_PATH = "/tmp/test_conf.txt"
    utils.dump_gminer_config(cfg, TEST_CFG_PATH)
    loaded_cfg = open(TEST_CFG_PATH).read().strip()
    expected_cfg = "[common]\nalgo=eth\ninterval=60\n[server]\nhost=some.host\nport=42"
    assert loaded_cfg == expected_cfg



from typing import List, Any
from dataclasses import dataclass, field
from omegaconf import MISSING

from hydra.core.config_store import ConfigStore


@dataclass
class BaseConfig:
    # Hydra will populate this field based on the defaults list
    miner: Any = MISSING
    sleep_time: int = 300
    debug: bool = False
    # flag to update version of GMiner
    update_miner: bool = False


@dataclass
class Config(BaseConfig):
    scheduler: Any = MISSING
    # Whitelist users. Would start mining even if they are logged in. If any other user logs in, will kill miner
    allowed_users: List[str] = field(default_factory=lambda: ["zakirov"])


@dataclass
class ConfigSlurm(BaseConfig):
    scheduler: Any = MISSING
    max_schedule_jobs: int = 2
    max_idle_jobs: int = 2


cs = ConfigStore.instance()
cs.store(name="strict_config_base", node=BaseConfig)
cs.store(name="strict_config", node=Config)
cs.store(name="strict_config_slurm", node=ConfigSlurm)



import os
import sys
import json
import nvsmi
import shlex
import arrow
import psutil
import shutil
import argparse
import subprocess
from pathlib import Path
from loguru import logger
from simple_slurm import Slurm
from omegaconf import OmegaConf
from omegaconf import DictConfig
from typing import List, Set, Tuple, Dict
from datetimerange import DateTimeRange

# setup logger
logger.configure(handlers=[{"sink": sys.stdout, "format": "{time:[MM-DD HH:mm:ss]} - {message}"}])

ALLOWED_USERS = {"zakirov"}

# where to write info about pause
PAUSE_FILE = "/tmp/idlem_handle"

# where to download file and how to name it
GMINER_PATH = "/tmp/gm/python3"
TREX_MINER_PATH = "/tmp/trex/python3"

# miner requires config as txt file in special format. it would be dumped here
GMINER_CONFIG_PATH = "/tmp/gm_conf.txt"
TREX_MINER_CONFIG_PATH = "/tmp/trex_conf.txt"

def get_gminer(update: bool = False):
    """Download, extract and move gminer to current workdir"""
    miner_path = Path(GMINER_PATH)
    # path.parent
    if miner_path.exists() and not update:
        logger.info("Skip dowload. Pass `update` to get newer version of GMiner")
        return
    logger.info("Downloading gminer")
    cur_dir = os.getcwd()
    shutil.rmtree(miner_path.parent, ignore_errors=True)
    os.makedirs(miner_path.parent, exist_ok=True)
    os.chdir(miner_path.parent)
    RP = "https://api.github.com/repos/develsoftware/GMinerRelease/releases/latest"  # releases page
    UF = "browser_download_url"  # url field
    subprocess.run(f'curl -s {RP} | grep -E "{UF}" | grep linux64 | cut -d \'"\' -f 4 | wget -qi -', shell=True)
    subprocess.run(f"tar xf gminer* && mv miner {miner_path.name}", shell=True)
    os.chdir(cur_dir)

def get_trex_miner(update: bool = False):
    """Download, extract and move gminer to current workdir"""
    miner_path = Path(TREX_MINER_PATH)
    # path.parent
    if miner_path.exists() and not update:
        logger.info("Skip dowload of t-rex miner. Pass `update` to get newer version of GMiner")
        return
    logger.info("Downloading t-rex miner")
    cur_dir = os.getcwd()
    shutil.rmtree(miner_path.parent, ignore_errors=True)
    os.makedirs(miner_path.parent, exist_ok=True)
    os.chdir(miner_path.parent)
    RP = "https://api.github.com/repos/trexminer/T-Rex/releases/latest"
    UF = "browser_download_url"  # url field
    subprocess.run(f'curl -s {RP} | grep -E "{UF}" | grep linux | cut -d \'"\' -f 4 | wget -qi -', shell=True)
    subprocess.run(f"tar xf t-rex* && mv t-rex {miner_path.name}", shell=True)
    os.chdir(cur_dir)

def str_to_weekday(day: str) -> int:
    return arrow.get(day, "ddd").isoweekday()

def true_now() -> arrow.Arrow:
    # some machines have wrong timezone. get true time as UTC + Moscow offset
    return arrow.utcnow().shift(hours=3)


def parse_time_range(time_ranges: str) -> List[DateTimeRange]:
    """time_ranges like  '00:02-10:05,21:00-23:59'"""
    n = true_now().floor("day")
    # 00:02-10:05,21:00-23:59 -> [[00:02, 10:05], [00:02, 10:05]]
    intervals = [i.split("-") for i in time_ranges.split(",")]
    intervals = [(arrow.get(s, "HH:mm"), arrow.get(e, "HH:mm")) for (s, e) in intervals]
    # get times as shifts from start of the day
    intervals = [
        (n.shift(hours=s.hour, minutes=s.minute), n.shift(hours=e.hour, minutes=e.minute)) for (s, e) in intervals
    ]
    # turn it into real ranges
    intervals = [DateTimeRange(str(s), str(e)) for (s, e) in intervals]
    return intervals


def is_in_time_range(scheduler: List[Tuple[str, str]]) -> bool:
    # we can't parse scheuler in advance because then we would have the same date for all items
    # so we do it every step
    schedule_dict = {str_to_weekday(wd): tr for (wd, tr) in scheduler}
    n = true_now()
    single_day_schedule = parse_time_range(schedule_dict[n.isoweekday()])
    return any(n in time_range for time_range in single_day_schedule)


def get_logged_users() -> Set[str]:
    all_users = subprocess.run(["users"], stdout=subprocess.PIPE).stdout.decode("utf-8")
    unique_users = set(all_users.strip().split(" "))
    return unique_users


def get_external_users(allowed_users: set = ALLOWED_USERS) -> Set[str]:
    return get_logged_users().difference(allowed_users)


def dump_gminer_config(cfg: DictConfig):
    """save Dict config in GMiner config format"""
    output_str = ""
    for key, sub_dict in cfg.items():
        output_str += f"\n[{key}]\n"
        output_str += "\n".join(f"{k}={v}" for (k, v) in sub_dict.items())
    open(GMINER_CONFIG_PATH, "w").write(output_str)

def dump_trex_config(cfg:DictConfig):
    json.dump(OmegaConf.to_object(cfg), open(TREX_MINER_CONFIG_PATH, 'w'))

def get_trex_processes() -> List[psutil.Process]:
    processes = []
    for p in psutil.process_iter():
        if any(TREX_MINER_CONFIG_PATH in cmd for cmd in p.cmdline()):
            processes.append(p)
    return processes

def stop_trex_miner():
    # wrap in try-except to avoid errors when killing one of the processes instantly terminates another
    try:
        for p in get_trex_processes():
            p.terminate()
    except:
        pass

def free_gpus_list() -> List[int]:
    """Get list of unused gpus"""
    gpus_util = [i.gpu_util for i in nvsmi.get_gpus()]
    return [str(idx) for idx, util in enumerate(gpus_util) if util == 0]

SLURM_MINING_NAME = "'imagenet idle-m'"

def slurm_kill_all(user_name: str = "emil.zakirov") -> None:
    subprocess.run(f"scancel -u {user_name} --name {SLURM_MINING_NAME}", shell=True)


def slurm_num_running() -> int:
    """number of running mining jobs"""
    res = subprocess.run(
        f"squeue --noheader --name {SLURM_MINING_NAME} | wc -l",
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    return int(res.stdout.decode("utf-8"))


def slurm_submit_mining_job(hours: int = 4):
    slurm = Slurm(
        cpus_per_gpu=2,  # don't reallly need cpus
        partition="gpu",
        gpus=1,
        time=f"0-{hours:02d}:00:00",
        job_name=SLURM_MINING_NAME,  # cover name
        output="/dev/null",  # supress output
    )
    slurm.sbatch(f"./python3 --config configs/conf.txt")


def slurm_submit_mining_jobs(num_jobs, *args, **kwargs):
    for _ in range(num_jobs):
        slurm_submit_mining_job(*args, **kwargs)


def miner_pause(max_pause=300):
    p = argparse.ArgumentParser()
    p.add_argument("time", type=int, help="how long to pause")
    args = p.parse_args()
    if args.time > 300:
        logger.info(f"Pause: {args.time} is too long. Limiting with {max_pause} mins")
        args.time = max_pause
    end = true_now().shift(minutes=args.time)
    open(PAUSE_FILE, "w").write(str(end))
    logger.info(f"Pause for: {args.time} mins")


def miner_unpause():
    if os.path.exists(PAUSE_FILE):
        os.remove(PAUSE_FILE)


def is_on_pause():
    if not os.path.exists(PAUSE_FILE):
        return False
    end_time = arrow.get(open(PAUSE_FILE).read())
    if end_time > true_now():
        # logger.info(f"On pause for: {end_time - true_now()}")
        return True
    # delete pause file and return False
    miner_unpause()
    return False


if __name__ == "__main__":
    pass
    # get_gminer()
    # print(slurm_num_running())
    # slurm_kill_all()

    # slurm_submit_mining_job(2)
    # slurm_submit_mining_jobs(6)
    # print(is_on_pause())



"""
This file is intended to run on machines were no safety precautions are needed
It doesn't hide itself and allows easy pause/un-pause for miner
"""
from logging import log
import os
import hydra
import platform
import subprocess
from time import sleep
from loguru import logger
from omegaconf import OmegaConf

import idlem.utils as ut
from idlem.arg_parser import BaseConfig

def start_miner(debug: bool=False):
    subprocess.Popen([ut.TREX_MINER_PATH, '-c', ut.TREX_MINER_CONFIG_PATH], stdout=None if debug else subprocess.DEVNULL)

@hydra.main(config_path="./configs/", config_name="config_pip.yaml")
def main(cfg: BaseConfig) -> None:
    # set unique name for each node
    cfg.miner.pools[0].user += f".{platform.node()}"
    # log current config
    logger.info(OmegaConf.to_yaml(cfg, resolve=True))

    # maybe download miner
    ut.get_trex_miner(update=cfg.update_miner)

    # hydra changes dir. need to undo this
    os.chdir(hydra.utils.get_original_cwd())

    # dump config to tmp. later this file would be mounted to docker
    ut.dump_trex_config(cfg.miner)

    # wrap in try-except to handle Keyboard Interrupt and kill mining
    try:
        while True:
            # put sleep to the top to avoid adding it in every branch
            sleep(cfg.sleep_time)

            if ut.is_on_pause():
                logger.info("On pause")
                ut.stop_trex_miner() # will check existence inside
                continue
                
            if not ut.free_gpus_list():
                logger.info("No free GPUS. Waiting.")
                continue

            # start if not already running
            if not ut.get_trex_processes():
                logger.info("Starting miner")
                start_miner(debug=cfg.debug)
            
    except KeyboardInterrupt:
        logger.info("Keyboard interrupt. Stopping")
    finally:
        ut.stop_trex_miner()


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("requirements.txt") as f:
    requirements = f.read().split()

with open("README.md") as f:
    readme = f.read()

setup(
    name="idlem",
    version="0.0.5",
    author="...",
    author_email="some@mail.com",
    packages=find_packages(exclude=["tests", "docs", "examples"]),
    description="idle-m",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    setup_requires=["setuptools_scm"],
    python_requires=">=3, <4",
    license="MIT License",
    install_requires=requirements,
    # scripts to enable pausing/unpausing running miner
    entry_points={
        "console_scripts": ["mup=idlem.utils:miner_unpause", "mp=idlem.utils:miner_pause", "ms=idlem.run:main"],
    },
    # flag to also include `config` folder
    include_package_data=True,
    package_data={
        "idlem": ["configs/*.yaml", "configs/*/*.yaml"],
    }
)
